# -*- coding: utf-8 -*-
"""
"""
import pytest
import types
import sys

from collections import Counter
from uhashring import HashRing
from uuid import uuid4


PY3 = sys.version_info >= (3,)


@pytest.fixture
def ring():
    ring = HashRing(
        nodes={'node1': 1,
               'node2': 1,
               'node3': 1},
        hash_fn='ketama')
    return ring


@pytest.fixture
def ring_fast():
    ring = HashRing(
        nodes={'node1': 1,
               'node2': 1,
               'node3': 1},
        compat=False)
    return ring


def weight_fn(**conf):
    print(conf)
    return int(conf['nodename'][-1])


def test_ring_size(ring, ring_fast):
    assert ring.size == len(ring.get_nodes()) * 160
    assert ring.size == len(ring._keys)
    assert ring.size == len(ring._ring)
    assert ring.size == len(ring.get_points())

    assert ring_fast.size == len(ring_fast.get_nodes()) * 160
    assert ring_fast.size == len(ring_fast._keys)
    assert ring_fast.size == len(ring_fast._ring)
    assert ring_fast.size == len(ring_fast.get_points())

    assert ring.size == ring_fast.size


def test_consistency(ring):
    coconut_node = ring.get('coconut')
    zoro_node = ring.get('zoro')

    del ring['node3']

    with pytest.raises(KeyError):
        del ring['node3']

    assert coconut_node == ring.get('coconut')
    assert zoro_node != ring.get('zoro')

    ring.add_node('node3', 1)

    assert zoro_node == ring.get('zoro')


def test_aliases(ring):
    assert ring.add_node == ring.__setitem__
    assert ring.get_node_instance == ring.__getitem__
    assert ring.remove_node == ring.__delitem__
    assert ring.continuum == ring.ring
    assert ring.nodes == ring.conf


def test_ketama_ring_hashi(ring):
    h = ring.hashi('test')
    assert h == 3446378249 == ring.get_key('test')


def test_faster_ring_hashi(ring_fast):
    h = ring_fast.hashi('test')
    assert h == 12707736894140473154801792860916528374


# XXX https://docs.python.org/3/whatsnew/3.0.html#ordering-comparisons
@pytest.mark.skipif(PY3, reason="requires python26, python27")
def test_range(ring):
    r = list(ring.range('test'))
    r.sort()
    c = ring.conf.values()
    c.sort()
    assert r == c

    r = list(ring.range('test', size=2, unique=True))
    assert len(r) == 2

    r = list(ring.range('test', size=2, unique=False))
    assert len(r) == 2

    r = list(ring.range('1800'))  # 1800 is at position 0
    r = list(ring.range('849'))  # 849 is at position -2

    r = list(ring.range('test', size=None, unique=False))
    assert len(r) == ring.size

    l = []
    for node in ring.iterate_nodes('test'):
        l.append(node)
    assert len(l) == 3

    ring_empty = HashRing()
    for node in ring_empty.iterate_nodes('test'):
        assert node is None


def test_conf():
    ring_1 = HashRing({'node1': 1})
    ring_2 = HashRing('node1')
    ring_3 = HashRing(['node1'])
    ring_4 = HashRing({'node1': {'weight': 1}})

    assert ring_1.ring == ring_2.ring == ring_3.ring == ring_4.ring

    with pytest.raises(ValueError):
        HashRing({'node1': 'fail'})

    with pytest.raises(ValueError):
        HashRing(None)


def test_empty_ring():
    ring = HashRing()
    assert ring.get('test') is None


def test_methods_return_types(ring):
    assert isinstance(ring['test'], type(None))
    assert isinstance(ring.get('test'), dict)
    assert isinstance(ring.get_instances(), list)
    assert isinstance(ring.get_node('test'), str)
    assert isinstance(ring.get_nodes(), type({}.keys()) if PY3 else list)
    assert isinstance(ring.get_node_hostname('test'), str)
    assert isinstance(ring.get_node_port('test'), type(None))
    assert isinstance(ring.get_node_pos('test'), int)
    assert isinstance(ring.get_node_weight('test'), int)
    assert isinstance(ring.get_points(), list)
    assert isinstance(ring.get_server('test'), tuple)
    assert isinstance(ring.iterate_nodes('test'), types.GeneratorType)
    assert isinstance(ring.range('test'), types.GeneratorType)
    assert isinstance(ring.distribution, Counter)


def test_print_without_error(ring):
    assert ring.print_continuum() is None
    ring = HashRing()
    assert ring.print_continuum() is None


def test_with_non_str_objects(ring):
    uid = uuid4()
    ring.get_node(uid)


def test_weight_fn():
    ring = HashRing(
        nodes={'node1': 1, 'node2': 1, 'node3': 1},
        replicas=4,
        vnodes=40,
        hash_fn='ketama',
        weight_fn=weight_fn)

    assert ring.distribution['node1'] == 80
    assert ring.distribution['node2'] == 160
    assert ring.distribution['node3'] == 240

    ring.regenerate

    assert ring.distribution['node1'] == 80
    assert ring.distribution['node2'] == 160
    assert ring.distribution['node3'] == 240

    with pytest.raises(TypeError):
        ring = HashRing(
            nodes={'node1': 1, 'node2': 1, 'node3': 1},
            replicas=4,
            vnodes=40,
            hash_fn='ketama',
            weight_fn=12)

    with pytest.raises(TypeError):
        ring = HashRing(
            nodes={'node1': 1, 'node2': 1, 'node3': 1},
            replicas=4,
            vnodes=40,
            hash_fn='ketama',
            weight_fn='coconut')


def test_ring_growth_ketama(ring):
    add_ring = HashRing(hash_fn='ketama')
    for nodename in ring.nodes:
        add_ring.add_node(nodename)

    assert ring._nodes == add_ring._nodes
    assert ring.ring == add_ring.ring
    assert ring.distribution == add_ring.distribution


def test_ring_growth_meta(ring_fast):
    add_ring = HashRing(compat=False)
    for nodename in ring_fast.nodes:
        add_ring.add_node(nodename)

    assert ring_fast._nodes == add_ring._nodes
    assert ring_fast.ring == add_ring.ring
    assert ring_fast.distribution == add_ring.distribution

def test_ketama_ring_shrink_collision():
    """
    see issue #6 thanks to @bjhockley
    """
    nodes = ["172.31.1.0", "172.31.1.125", "172.31.1.202"]
    ring = HashRing(nodes, hash_fn='ketama')
    ring.remove_node(nodes[1])
    ring.remove_node(nodes[2])
    ring.remove_node(nodes[0])
    assert ring.ring == {}



# -*- coding: utf-8 -*-
"""This is not part of the test suite.
"""
try:
    import ketama
except Exception:
    ketama = None

from time import time
from tempfile import NamedTemporaryFile
from uhashring import HashRing

num = 1000000
print('running {} key generation comparison'.format(num))

# ketama C binding
if ketama:
    with NamedTemporaryFile(prefix='benchmark_') as ketama_config_file:
        ketama_config_file.write("127.0.0.1:11211\t600\n")
        ketama_config_file.write("127.0.0.1:11212\t400\n")
        ketama_config_file.flush()

        kt = ketama.Continuum(ketama_config_file.name)
        pt = time()
        for i in range(num):
            key = 'myval-{}'.format(i)
            kt.get_server(key)
        print('ketama took {} s'.format(time() - pt))

# pure python implementation
ring = HashRing(
    nodes={'127.0.0.1:11211': 600,
           '127.0.0.1:11212': 400},
    replicas=4,
    vnodes=40,
    compat=True)
pt = time()
for i in range(num):
    key = 'myval-{}'.format(i)
    ring.get_server(key)
print('HashRing took {} s'.format(time() - pt))



"""
uhashring
"""

import os
import sys

from pkg_resources import get_distribution, parse_version
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)

        # https://bitbucket.org/pypa/setuptools/commits/cf565b6
        if get_distribution('setuptools').parsed_version < parse_version('18.4'):
            self.test_args = []
            self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

setup(
    author='Ultrabug',
    author_email='ultrabug@ultrabug.net',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    cmdclass={'test': PyTest},
    description='Full featured consistent hashing python library compatible with ketama.',
    download_url='https://github.com/ultrabug/uhashring/tags',
    include_package_data=True,
    install_requires=[],
    license='BSD',
    long_description=read('README.rst'),
    name='uhashring',
    packages=find_packages(),
    platforms='any',
    tests_require=['python-memcached', 'pytest'],
    url='https://github.com/ultrabug/uhashring',
    version='1.0',
    zip_safe=True)



# -*- coding: utf-8 -*-

from collections import Counter
from hashlib import md5


class MetaRing(object):
    """Implement a tunable consistent hashing ring."""

    def __init__(self, hash_fn):
        """Create a new HashRing.

        :param hash_fn: use this callable function to hash keys.
        """
        self._distribution = Counter()
        self._keys = []
        self._nodes = {}
        self._ring = {}

        if hash_fn and not hasattr(hash_fn, '__call__'):
            raise TypeError('hash_fn should be a callable function')
        self._hash_fn = hash_fn

    def hashi(self, key):
        """Returns an integer derived from the md5 hash of the given key.
        """
        return int(md5(str(key).encode('utf-8')).hexdigest(), 16)

    def _create_ring(self, nodes):
        """Generate a ketama compatible continuum/ring.
        """
        for node_name, node_conf in nodes:
            for w in range(0, node_conf['vnodes'] * node_conf['weight']):
                self._distribution[node_name] += 1
                self._ring[self.hashi('%s-%s' % (node_name, w))] = node_name
        self._keys = sorted(self._ring.keys())

    def _remove_node(self, node_name):
        """Remove the given node from the continuum/ring.

        :param node_name: the node name.
        """
        try:
            node_conf = self._nodes.pop(node_name)
        except:
            raise KeyError('node \'{}\' not found, available nodes: {}'.format(
                node_name, self._nodes.keys()))
        else:
            self._distribution.pop(node_name)
            for w in range(0, node_conf['vnodes'] * node_conf['weight']):
                del self._ring[self.hashi('%s-%s' % (node_name, w))]
            self._keys = sorted(self._ring.keys())
